/*
 *  TINET (TCP/IP Protocol Stack)
 * 
 *  Copyright (C) 2001-2017 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡξȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: ppp_ipv6cp.c 1.7 2017/6/1 8:49:7 abe $
 */

/*
 * ipv6cp.c - PPP IP Control Protocol.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 *	PPP IP Control Protocol (IPV6CP) Module
 *
 *	    Written by Toshiharu OHNO (tony-o@iij.ad.jp)
 *
 *   Copyright (C) 1993, Internet Initiative Japan, Inc. All rights reserverd.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Internet Initiative Japan, Inc.  The name of the
 * IIJ may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $FreeBSD: src/usr.sbin/ppp/ipv6cp.c,v 1.90.2.4 2000/08/19 09:30:03 brian Exp $
 *
 *	TODO:
 *		o Support IPADDRS properly
 *		o Validate the length in IpcpDecodeConfig
 */

#include <string.h>

#include <s_services.h>
#include <t_services.h>
#include "kernel_id.h"

#include <tinet_defs.h>
#include <tinet_config.h>

#include <net/if.h>
#include <net/if_ppp.h>
#include <net/net.h>
#include <net/net_endian.h>
#include <net/net_buf.h>
#include <net/net_count.h>
#include <net/ppp.h>
#include <net/ppp_var.h>
#include <net/ppp_fsm.h>
#include <net/ppp_lcp.h>
#include <net/ppp_ipv6cp.h>
#include <net/ppp_modem.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/tcp_var.h>
#include <netinet/tcp_timer.h>
#include <netinet/udp_var.h>

#include <net/if_var.h>

#ifdef SUPPORT_PPP

/*
 *  ؿ
 */

static void ipv6cp_init (void);
static void ipv6cp_input (T_NET_BUF *input);
static void ipv6cp_protrej (void);
static void ipv6cp_lowerup (void);
static void ipv6cp_lowerdown (void);
static void ipv6cp_open (void);
static void ipv6cp_close (void);

/*
 * FSM ƽФ륳Хå
 */

static void  ipv6cp_resetci(T_PPP_FSM *fsm);	/* ꥻåȤ롣	*/
static int_t ipv6cp_cilen	(T_PPP_FSM *fsm);	/* Ĺ֤		*/
static void  ipv6cp_addci	(T_PPP_FSM *fsm, T_NET_BUF *output);
						/* ɲä롣		*/
static bool_t  ipv6cp_ackci	(T_PPP_FSM *fsm, T_NET_BUF *input);
						/* ACK Ȥν	*/
static bool_t  ipv6cp_nakci	(T_PPP_FSM *fsm, T_NET_BUF *input);
						/* NAK Ȥν	*/
static bool_t  ipv6cp_rejci	(T_PPP_FSM *fsm, T_NET_BUF *input);
						/* REJ Ȥν	*/
static int_t ipv6cp_reqci	(T_PPP_FSM *fsm, T_NET_BUF *input, T_NET_BUF *output);
						/* ιǧ롣	*/
static void ipv6cp_up	(T_PPP_FSM *fsm);	/* ³Ω롣		*/
static void ipv6cp_down	(T_PPP_FSM *fsm);	/* ³롣		*/
static void ipv6cp_finished(T_PPP_FSM *fsm);	/* ؤλ롣		*/

/*
 *  ѿ
 */

static T_IFNET ipv6cp_local_def_cfg = {		/* ε		*/
	NULL,					/* ͥåȥ󥿥եΥեȥ	*/
	{},					/* IPv6 ɥ쥹				*/
	{},					/* ޥ㥹 IPv6 ɥ쥹			*/
	0,					/* ե饰					*/
	};

static T_IFNET ipv6cp_remote_def_cfg = {	/* ιε	*/
	NULL,					/* ͥåȥ󥿥եΥեȥ	*/
	{},					/* IPv6 ɥ쥹				*/
	{},					/* ޥ㥹 IPv6 ɥ쥹			*/
	0,					/* ե饰					*/
	};

static T_IFNET ipv6cp_local_neg_cfg;		/* ͥμ		*/

static T_IFNET ipv6cp_remote_neg_cfg;		/* ͥι		*/

/*
 *  ѿ
 */

T_PPP_FSM_CALLBACKS ipv6cp_callbacks = {
	ipv6cp_resetci,			/* ꥻåȤ롣	*/
	ipv6cp_cilen,			/* Ĺ֤		*/
	ipv6cp_addci,			/* ɲä롣		*/
	ipv6cp_ackci,			/* ACK Ȥν	*/
	ipv6cp_nakci,			/* NAK Ȥν	*/
	ipv6cp_rejci,			/* REJ Ȥν	*/
	ipv6cp_reqci,			/* ιǧ롣	*/
	ipv6cp_up,			/* ³Ω롣		*/
	ipv6cp_down,			/* ³롣		*/
	NULL,				/* ؤ򳫻Ϥ롣		*/
	ipv6cp_finished,			/* ؤλ롣		*/
	NULL,				/* 롣			*/
	NULL,				/*  CP Ȥν	*/
	};

T_PPP_FSM ipv6cp_fsm = {
	&ipv6cp_callbacks,		/* Хåؿ		*/
	PPP_IPV6CP,			/* ץȥ			*/
	};

T_PPP_PROTENT ipv6cp_protent = {
	PPP_IPV6CP,
	ipv6cp_init,			/* 			*/
	ipv6cp_input,			/* 				*/
	ipv6cp_protrej,			/* Proto-REJ 		*/
	ipv6cp_lowerup,			/* ؤư		*/
	ipv6cp_lowerdown,		/* ؤߤ		*/
	ipv6cp_open,			/* ץ󤹤			*/
	ipv6cp_close,			/* 			*/
	ip6_input,			/* ǡϡIPv6 	*/
	};

T_IFNET ipv6cp_local_ack_cfg;		/* ˵Ĥ줿μ	*/
T_IFNET ipv6cp_remote_ack_cfg;		/* ˵Ĥ	*/

/*
 *  ppp_get_ifnet -- ͥåȥ󥿥ե¤Τ֤
 */

T_IFNET *
ppp_get_ifnet (void)
{
	return &ipv6cp_local_ack_cfg;
	}

/*
 *  wait_ipv6cp -- IP ³λޤԤġ
 *
 *	: ³˼Ԥ E_OBJ ֤
 */

ER
wait_ipv6cp (void)
{
#ifdef PPP_CFG_MODEM
	ER error;
#endif	/* of #ifdef PPP_CFG_MODEM */

	if (ipv6cp_fsm.state <= PPP_FSM_STOPPED) {

#ifdef PPP_CFG_MODEM

		/* ǥ³λޤԤġ*/
		if ((error = wait_modem()) != E_OK)
			return error;

#else	/* of #ifdef PPP_CFG_MODEM */

		/*
		 * ưץ󤷡꤫³Ԥʤ LCP ٥
		 * ǽưץ֤ǡƥץ󤹤롣
		 */
		lcp_close();
		lcp_open(PPP_OPEN_ACTIVE);

#endif	/* of #ifdef PPP_CFG_MODEM */

		}

	if (ipv6cp_fsm.state != PPP_FSM_OPENED) {
		/* IP ³λޤԤġ*/
		wai_sem(SEM_IPV6CP_READY);
		if (ipv6cp_fsm.state != PPP_FSM_OPENED)
			return E_OBJ;
		}

	return E_OK;
	}

/*
 *  ipv6cp_init -- IPV6CP ⥸塼ν
 */

static void
ipv6cp_init (void)
{
	fsm_init(&ipv6cp_fsm);
	}

/*
 *  ipv6cp_input -- IPV6CP 
 */

static void
ipv6cp_input (T_NET_BUF *input)
{
	NET_COUNT_PPP_IPV6CP(net_count_ppp_ipv6cp_in_octets,  input->len);
	NET_COUNT_PPP_IPV6CP(net_count_ppp_ipv6cp_in_packets, 1);
	fsm_input(&ipv6cp_fsm, input);
	}

/*
 *  ipv6cp_resetci -- IPV6CPιꥻåȤ롣
 */

static void
ipv6cp_resetci (T_PPP_FSM *fsm)
{
	/* ν */
	ipv6cp_local_neg_cfg = ipv6cp_local_def_cfg;
	memset((void*)&ipv6cp_local_ack_cfg, 0, sizeof(T_IFNET));

	/* ιν */
	ipv6cp_remote_neg_cfg = ipv6cp_remote_def_cfg;
	memset((void*)&ipv6cp_remote_ack_cfg, 0, sizeof(T_IFNET));
	}

/*
 *  ipv6cp_cilen -- Ĺ֤
 */

static int_t
ipv6cp_cilen (T_PPP_FSM *fsm)
{
	int_t cilen = 0;

	cilen += sizeof(T_PPP_CI_HDR) + sizeof(uint32_t);	/* IPv6 ɥ쥹 */

	return cilen;
	}

/*
 *  ipv6cp_addci -- IPV6CPιɲä롣
 */

static void
ipv6cp_addci (T_PPP_FSM *fsm, T_NET_BUF *output)
{
	uint8_t *buf;

	buf = output->buf + sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);

	/* IPv6 ɥ쥹ץɲä롣 */
	*buf ++ = IPV6CP_CIT_ADDR;
	*buf ++ = sizeof(T_PPP_CI_HDR) + sizeof(uint32_t);
	ahtonl(buf, ipv6cp_local_neg_cfg.in6_ifaddrs[0].addr);
	buf += sizeof(uint32_t);
	}

/*
 *  ipv6cp_ackci -- ACK Ȥν
 *
 *	:	true  ACK 
 *		false ACK ϰ۾
 */

static bool_t
ipv6cp_ackci (T_PPP_FSM *fsm, T_NET_BUF *input)
{
	uint8_t		*buf;
	uint16_t	len;

	buf = input->buf +  sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);
	len = input->len - (sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR));

	/*
	 *  ˴˰פʤФʤʤ
	 *  äơipv6cp_addci Ǻ˲Ϥ롣
	 */

	/* ADDR ץϤ롣 */
	if (len < sizeof(T_PPP_CI_HDR) + sizeof(uint32_t)	||
	    *buf ++ != IPV6CP_CIT_ADDR			||
	    *buf ++ != sizeof(T_PPP_CI_HDR) + sizeof(uint32_t)||
	    nahcmpl(buf, ipv6cp_local_neg_cfg.in6_ifaddrs[0].addr))
		return false;
	buf += sizeof(uint32_t);
	len -= sizeof(T_PPP_CI_HDR) + sizeof(uint32_t);

	return true;
	}

/*
 *  ipv6cp_nakci -- NAK Ȥν
 *
 *	:	1  NAK 
 *		0  NAK ϰ۾
 */

static int_t
ipv6cp_nakci (T_PPP_FSM *fsm, T_NET_BUF *input)
{
	T_IFNET		new_cfg;
	uint32_t	addr;
	uint16_t	len;
	uint8_t		*buf, cilen;

	buf = input->buf +  sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);
	len = input->len - (sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR));

	new_cfg = ipv6cp_local_neg_cfg;

	/* ADDR ץϤ롣 */
	if (len >= sizeof(T_PPP_CI_HDR) + sizeof(addr)		&&
	    * buf      == IPV6CP_CIT_ADDR			&&
	    *(buf + 1) == sizeof(T_PPP_CI_HDR) + sizeof(addr)) {
		memcpy(&new_cfg.in6_ifaddrs[0].addr, buf + 2, sizeof(addr));
		buf += sizeof(T_PPP_CI_HDR) + sizeof(addr);
		len -= sizeof(T_PPP_CI_HDR) + sizeof(addr);
		}

	/* ̵뤹롣*/
	while (len > sizeof(T_PPP_CI_HDR)) {
		cilen = *(buf + 1);
		if (len < cilen)
			return 0;
		buf += cilen;
		len -= cilen;
		}

	/* Ĺ 0 ǤʤХ顼 */
	if (len != 0)
		return 0;

	/* ץꤹ롣*/
	if (fsm->state != PPP_FSM_OPENED)
		ipv6cp_local_neg_cfg = new_cfg;
               		
	return 1;
	}

/*
 *  ipv6cp_rejci -- REJ Ȥν
 *
 *	:	1  REJ 
 *		0  REJ ϰ۾
 */

static int_t
ipv6cp_rejci (T_PPP_FSM *fsm, T_NET_BUF *input)
{
	T_IFNET		new_cfg;
	uint8_t		*buf;
	uint16_t	len;

	buf = input->buf +  sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);
	len = input->len - (sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR));

	new_cfg = ipv6cp_local_neg_cfg;

	/* ADDR ץϤ롣 */
	if (len >= sizeof(T_PPP_CI_HDR) + sizeof(new_cfg.in6_ifaddrs[0].addr)	&&
	    * buf      == IPV6CP_CIT_ADDR					&&
	    *(buf + 1) == sizeof(T_PPP_CI_HDR) + sizeof(new_cfg.in6_ifaddrs[0].addr)) {
		memcpy(&new_cfg.in6_ifaddrs[0].addr, &in6_addr_unspecified,
		 sizeof(new_cfg.in6_ifaddrs[0].addr));
		buf += sizeof(T_PPP_CI_HDR) + sizeof(new_cfg.in6_ifaddrs[0].addr);
		len -= sizeof(T_PPP_CI_HDR) + sizeof(new_cfg.in6_ifaddrs[0].addr);
		}

	/* Ĺ 0 ǤʤХ顼 */
	if (len != 0)
		return 0;

	/* ץꤹ롣*/
	if (fsm->state != PPP_FSM_OPENED)
		ipv6cp_local_neg_cfg = new_cfg;

	return 1;
	}

/*
 *  ipv6cp_reqci -- ιϤ롣
 */

static int_t
ipv6cp_reqci (T_PPP_FSM *fsm, T_NET_BUF *input, T_NET_BUF *output)
{
	T_IN6_ADDR	addr;
	uint16_t	code, final, len, cilen;
	uint8_t		*np, *rp, *ap, *cp, type = 0;

	rp = ap = input->buf +  sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);
				/* rp: REJap: ACK  CI Υݥ	*/
				/* ݤϡ buf ˾񤭤롣	*/
	len   = input->len - (sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR));
	np    = output->buf + sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR);
	final = PPP_CONFACK;	/* ǽŪ륳 */
	while (len > 0) {
		code = PPP_CONFACK;
		cp   = ap;	/* cp: ߽ CI	*/

		if (len < sizeof(T_PPP_CI_HDR)) {

			/* Ĥ꤬ CI إåû */
			syslog(LOG_NOTICE, "[PPP/IPV6CP] bad CI len: %d.", len);
			cilen = len;	/* ǡĹ۾ξν	*/
			len  = 0;	/* 롼פȴ롣		*/
			code  = PPP_CONFREJ;
			goto endswitch;
			}

		type  = *ap ++;	/* CI η		*/
		cilen = *ap ++;	/* CI Ĺ		*/
		if (len < cilen) {

			/* Ĥ꤬ CI Ĺû */
			syslog(LOG_NOTICE, "[PPP/IPV6CP] bad CI len: %d.", cilen);
			cilen = len;	/* ǡĹ۾ξν	*/
			len  = 0;	/* 롼פȴ롣		*/
			code  = PPP_CONFREJ;
			goto endswitch;
			}
		len -= sizeof(T_PPP_CI_HDR);

		/* CI ηˤʬ롣*/
		switch (type) {

		case IPV6CP_CIT_ADDR:	/* IPv6 ɥ쥹 */

		 	/* CI Ĺإå + 4 ƥåȤǤʤХ顼 */
		 	if (cilen != sizeof(T_PPP_CI_HDR) + sizeof(addr))
				code = PPP_CONFREJ;
			else {
				memcpy((uint8_t*)&ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr, ap,
				           sizeof(ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr));
				if (!IN6_ARE_ADDR_EQUAL(&addr, &ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr) &&
				    (IN6_IS_ADDR_UNSPECIFIED(&addr) ||
				     IN6_IS_ADDR_UNSPECIFIED(&ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr))) {
					memcpy(np, (uint8_t*)&ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr,
					               sizeof(ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr));
					*np ++ = IPV6CP_CIT_ADDR;
					*np ++ = sizeof(T_PPP_CI_HDR) + sizeof(uint32_t);
					np    += sizeof(ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr);
					code = PPP_CONFNAK;
					}
#if 1	/* ׳ǧ */
				else if (IN6_IS_ADDR_UNSPECIFIED(&addr) &&
				         IN6_IS_ADDR_UNSPECIFIED(&ipv6cp_remote_neg_cfg.in6_ifaddrs[0]))
#else
				else if (IN6_IS_ADDR_UNSPECIFIED(&addr) ||
				         IN6_IS_ADDR_UNSPECIFIED(&ipv6cp_remote_neg_cfg.in6_ifaddrs[0]))
#endif
					code = PPP_CONFREJ;
				else
					ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr = addr;
				}
			break;

		default:
			syslog(LOG_INFO, "[PPP/IPV6CP] unexp opt: %d.", type);
			code = PPP_CONFREJ;
			break;
			}
endswitch:

		if (code == PPP_CONFNAK) {
			/* CI η Magic Number ξơNAK β	*/
			/*  MAX_PPP_FAILURES  Ķݤ롣	*/
			if (fsm->failures >= MAX_PPP_FAILURES)
				code = PPP_CONFREJ;
			else if (final != PPP_CONFREJ)
				final = PPP_CONFNAK;
			}

		/*  CI ݤ */
		if (code == PPP_CONFREJ) {
			if (rp < cp)		/* ˵ͤ롣*/
				memcpy(rp, cp, cilen);
			rp += cilen;
			final = PPP_CONFREJ;
			}

		ap   += cilen - sizeof(T_PPP_CI_HDR);
		len  -= cilen - sizeof(T_PPP_CI_HDR);
		}

	/* ǽŪĹĴ롣*/
	switch (final) {
	case PPP_CONFNAK:
		output->len = np - output->buf;
		memcpy(output->buf, input->buf, sizeof(T_PPP_HDR) + sizeof(T_PPP_CP_HDR));
		break;
	case PPP_CONFREJ:
		memcpy(output->buf, input->buf, output->len = rp - input->buf);
		break;
	case PPP_CONFACK:
		memcpy(output->buf, input->buf, input->len);
		break;
		}

	return final;
	}

/*
 *  ipv6cp_up -- ³Ω롣
 */

static void
ipv6cp_up (T_PPP_FSM *fsm)
{
	/* ν */
	ipv6cp_local_ack_cfg = ipv6cp_local_neg_cfg;

	sig_sem(SEM_IPV6CP_READY);

	syslog(LOG_NOTICE, "[PPP/IPV6CP] up, Local IPv6 Addr: %s, Remote IPv6 Addr: %s.",
	                   ipv62str(NULL, &ipv6cp_local_neg_cfg.in6_ifaddrs[0].addr),
	                   ipv62str(NULL, &ipv6cp_remote_neg_cfg.in6_ifaddrs[0].addr));
	}

/*
 *  ipv6cp_down -- ³롣
 */

static void
ipv6cp_down (T_PPP_FSM *fsm)
{
	sig_sem(SEM_IPV6CP_READY);
	syslog(LOG_NOTICE, "[PPP/IPV6CP] down.");
	}

/*
 *  ipv6cp_finished -- ؤλ롣
 */

static void
ipv6cp_finished (T_PPP_FSM *fsm)
{
	}

/*
 *  ipv6cp_protrej -- Proto-REJ Ȥν
 */

static void
ipv6cp_protrej (void)
{
	fsm_lowerdown(&ipv6cp_fsm);
	}

/*
 *  ipv6cp_lowerup -- IPV6CP ؤư롣
 */

static void
ipv6cp_lowerup (void)
{
	fsm_lowerup(&ipv6cp_fsm);
	}

/*
 *  ipv6cp_lowerdown -- IPV6CP ؤߤ롣
 */

static void
ipv6cp_lowerdown (void)
{
	fsm_lowerdown(&ipv6cp_fsm);
	}

/*
 *  ipv6cp_open -- IPV6CP Υץ
 */

static void
ipv6cp_open (void)
{
	fsm_open(&ipv6cp_fsm);
	}

/*
 *  ipv6cp_close -- IPV6CP Υ
 */

static void
ipv6cp_close (void)
{
	fsm_close(&ipv6cp_fsm);
	}

#endif	/* fo #ifdef SUPPORT_PPP */
