/*
 * Copyright 2004 Henning Brauer <henning@openbsd.org>
 * Copyright (c) 1995, 1996, 1997, 1998, 1999
 * The Internet Software Consortium.    All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of The Internet Software Consortium nor the names
 *    of its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INTERNET SOFTWARE CONSORTIUM AND
 * CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE INTERNET SOFTWARE CONSORTIUM OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This software has been written for the Internet Software Consortium
 * by Ted Lemon <mellon@fugue.com> in cooperation with Vixie
 * Enterprises.  To learn more about the Internet Software Consortium,
 * see ``http://www.vix.com/isc''.  To learn more about Vixie
 * Enterprises, see ``http://www.vix.com''.
 *
 * This client was substantially modified and enhanced by Elliot Poger
 * for use on Linux while he was working on the MosquitoNet project at
 * Stanford.
 *
 * The current version owes much to Elliot's Linux enhancements, but
 * was substantially reorganized and partially rewritten by Ted Lemon
 * so as to use the same networking framework that the Internet Software
 * Consortium DHCP server uses.   Much system-specific configuration code
 * was moved into a shell script so that as support for more operating
 * systems is added, it will not be necessary to port and maintain
 * system-specific configuration code to these operating systems - instead,
 * the shell script can invoke the native tools to accomplish the same
 * purpose.
 */
/*
 * WIDE Project DHCP Implementation
 * Copyright (c) 1995-1997 Akihiro Tominaga
 * Copyright (c) 1995-1997 WIDE Project
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided the following conditions
 * are satisfied,
 *
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by WIDE Project and
 *      its contributors.
 * 3. Neither the name of WIDE Project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND WIDE
 * PROJECT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE. ALSO, THERE
 * IS NO WARRANTY IMPLIED OR OTHERWISE, NOR IS SUPPORT PROVIDED.
 *
 * Feedback of the results generated from any improvements or
 * extensions made to this software would be much appreciated.
 * Any such feedback should be sent to:
 * 
 *  Akihiro Tominaga
 *  WIDE Project
 *  Keio University, Endo 5322, Kanagawa, Japan
 *  (E-mail: dhcp-dist@wide.ad.jp)
 *
 * WIDE project has the rights to redistribute these changes.
 */
/*
 *  TINET (TCP/IP Protocol Stack)
 * 
 *  Copyright (C) 2011 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: dhcp4_cli.h 1.7 2017/6/1 8:50:19 abe $
 */

#ifndef DHCP4_CLI_H
#define DHCP4_CLI_H

#include <netapp/dhcp4.h>

/* 
 *  DHCP 饤ȤΥå
 */

#define DHCP4_CLI_STACK_SIZE		1024		/* DHCP 饤ȥΥå	*/

/* 
 *  DHCP 饤Ȥͥ
 */

#define DHCP4_CLI_MAIN_PRIORITY		5		/* DHCP 饤ȥͥ		*/

/* 侩٤ */
#define DHCP4_CLI_CFG_PREFER_ADDR	100		/* ɥ쥹ο侩	*/

/* DHCP 饤ȥåμꥹȥ2ʾꤹ뤳ȡ */
#define NUM_DHCP4_MSG_LIST		3

/*
 *  DHCP 饤ȥå¤οѹػߡ
 *  +1 
 */
#define NUM_MPF_DHCP4_CLI_MSG		(NUM_DHCP4_MSG_LIST+1)

/* 
 *  ѥե졼
 */

/* DHCP ɬܥץꥹ */

#ifndef DHCP4_CLI_CFG_REQUIRED_OLIST

#define DHCP4_CLI_CFG_REQUIRED_OLIST {	\
	DHCP4_OPT_SUBNET_MASK,		\
	DHCP4_OPT_ROUTER,		\
}

#endif	/* of #ifndef DHCP4_CLI_CFG_REQUIRED_OLIST */

/* DHCP ׵ᥪץꥹ */

#ifndef DHCP4_CLI_CFG_REQUEST_OLIST

#if defined(USE_RESOLVER)

#define DHCP4_CLI_CFG_REQUEST_OLIST {	\
	DHCP4_OPT_DNS_SERVER,		\
	DHCP4_OPT_DOMAIN_NAME,		\
}

#endif	/* of #if defined(USE_RESOLVER) */

#endif	/* of #ifndef DHCP4_CLI_CFG_REQUEST_OLIST */

/*
 *  ॢ
 */

/* ॢ	*/
#define TMO_DHCP4C_OFFER	ULONG_C(120*1000)	/* OFFER å	*/
#define TMO_DHCP4C_ACK		ULONG_C(60*1000)	/* ACK/NAK å	*/

/* ꡼֤δ	*/
#define TMO_DHCP4C_DEF_LEASE	ULONG_C(12*60*60*1000)	/* ɸ	*/
#define TMO_DHCP4C_MIN_LEASE	ULONG_C(60*1000)	/* Ǿ	*/

/* å	*/
#define TMO_DHCP4C_INIT_INTVL	ULONG_C(3*1000)		/* 	*/
#define TMO_DHCP4C_MAX_BACKOFF	ULONG_C(15*1000)	/* 	*/

/* DHCP Фؤ UDP ॢ	*/
#define TMO_DHCP4C_UDP_SND	ULONG_C(1000)

/* å¤γॢ	*/
#define TMO_DHCP4C_MPF_GET	ULONG_C(1000)

/* ARP ׵ƤԤ	*/
#define TMO_DHCP4C_ARP_WAIT	ULONG_C(1000)

/* RELEASE åƤԤ	*/
#define TMO_DHCP4C_SND_REL_WAIT	ULONG_C(1000)

#ifndef _MACRO_ONLY

#ifdef T_IPV4EP_DEFINED

/*
 *  DHCP 饤ȥå¤
 */

typedef struct dhcp4_cli_msg {
	/* 4Хȶ˥饤󤹤롣*/
	T_DHCP4_MSG		msg;		/* DHCP å¤	*/
	struct dhcp4_cli_msg	*next;		/* ꥹȤμΥå¤	*/
	T_IPV4EP		srv;		/* ФIPv4ɥ쥹/ݡֹ */
	uint16_t		len;		/* msg ͭĹ			*/
	uint16_t		prefer;		/* msg ο侩			*/
	uint8_t			type;		/* å		*/
	} T_DHCP4_CLI_MSG;

#endif	/*of #ifdef T_IPV4EP_DEFINED */

/*
 *  Хåؿ
 */

extern ER callback_nblk_dhcp4_cli (ID cepid, FN fncd, void *p_parblk);

/*
 *  
 */

extern void	dhcp4_cli_task(intptr_t exinf);

/*
 *  ؿ
 */

#ifdef T_IN4_ADDR_DEFINED

extern ER dhcp4c_rel_info (void);
extern ER dhcp4c_renew_info (void);
extern ER dhcp4c_get_info (T_IN4_ADDR *addr, uint32_t *expire, uint32_t *renew,
                                             uint32_t *rebind, SYSTIM *bind_start);

#endif	/* of #ifdef T_IN4_ADDR_DEFINED */

#endif	/* of #ifndef _MACRO_ONLY */

#endif	/* of #ifndef DHCP4_CLI_H */
