/*
 *  TINET (TCP/IP Protocol Stack)
 * 
 *  Copyright (C) 2001-2017 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: dns.h 1.7 2017/6/1 8:50:21 abe $
 */

#ifndef DNS_H
#define DNS_H

/*
 *  DNS å˴ؤ
 *
 *    RFC1034RFC1035RFC1886RFC2136
 *    RFC2671RFC2845RFC2930RFC3596
 *
 *  DNS å
 *
 *  +---------------------+
 *  |       Header        |
 *  +---------------------+
 *  |      Question       | Question section 
 *  +---------------------+
 *  |       Answer        | RR (Resouce Record) 
 *  +---------------------+
 *  |     Authority       | RR (Resouce Record) 
 *  +---------------------+
 *  |     Additional      | RR (Resouce Record) 
 *  +---------------------+
 *
 *  DNS إå
 *
 *    0                    7  8                   15
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                      ID                       |
 *  +--+-----------+--+--+--+--+--------+-----------+
 *  |QR|   Opcode  |AA|TC|RD|RA|    Z   |   RCODE   |
 *  +--+--------+--+--+--+--+--+--------+-----------+
 *  |                    QDCOUNT                    |
 *  +-----------------------------------------------+
 *  |                    ANCOUNT                    |
 *  +-----------------------------------------------+
 *  |                    NSCOUNT                    |
 *  +-----------------------------------------------+
 *  |                    ARCOUNT                    |
 *  +-----------------------------------------------+
 */

typedef struct dns_hdr {
	uint16_t	id;		/* ID */
	uint16_t	code;	/* QR:		1
				   Opcode:	4
				   AA:		1
				   TC:		1
				   RD:		1
				   RA:		1
				   Z:		3
				   RCODE:	4
				   */
	uint16_t	qdcount;	/* question   section */
	uint16_t	ancount;	/* answer     section */
	uint16_t	nscount;	/* authority  section */
	uint16_t	arcount;	/* additional section */
	} T_DNS_HDR;

/*
 *  code 
 */

/* QR  */

#define DNS_QR_QUERY		0x0000U
#define DNS_QR_RESPONSE		0x8000U

/* Opcode  */

#define DNS_OPCODE_MASK		0x7800U
#define DNS_OPCODE_FORWARD	0x0000U
#define DNS_OPCODE_REVERSE	0x0800U
#define DNS_OPCODE_STATUS	0x1000U

/* AA  */

#define DNS_AUTHORITATIVE	0x0400U

/* TC  */

#define DNS_TRUN_CATION		0x0200U

/* RD  */

#define DNS_RECURSION_DESIRED	0x0100U

/* RA  */

#define DNS_RECURSION_AVAILABLE	0x0080U

/* RCODE  */

#define DNS_RCODE_MASK		0x000fU
#define DNS_RCODE_NO_ERROR	0x0000U
#define DNS_RCODE_FORMAT_ERROR	0x0001U
#define DNS_RCODE_SERVER_ERROR	0x0002U
#define DNS_RCODE_NAME_ERROR	0x0003U
#define DNS_RCODE_NOT_IMPLEMENT	0x0004U
#define DNS_RCODE_REFUSED	0x0005U

/*
 * DNS TYPE 
 */

#define DNS_TYPE_A		(uint16_t)1U
#define DNS_TYPE_NS		(uint16_t)2U
#define DNS_TYPE_CNAME		(uint16_t)5U
#define DNS_TYPE_SOA		(uint16_t)6U
#define DNS_TYPE_PTR		(uint16_t)12U
#define DNS_TYPE_AAAA		(uint16_t)28U

/*
 * DNS CLASS 
 */

#define DNS_CLASS_IN		(uint16_t)1U

/*
 *  RR (Resouce Record) 
 *
 *    0                    7  8                   15
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                       .                       |
 *  +-----------------      .      -----------------+
 *  |                      NAME                     |
 *  +-----------------      .      -----------------+
 *  |                       .                       |
 *  +-----------------------------------------------+
 *  |                      TYPE                     |
 *  +-----------------------------------------------+
 *  |                      CLASS                    |
 *  +-----------------------------------------------+
 *  |                       TTL                     |
 *  |                                               |
 *  +-----------------------------------------------+
 *  |                    RDLENGTH                   |
 *  +-----------------------------------------------+
 *  |                       .                       |
 *  +-----------------      .      -----------------+
 *  |                     RDATA                     |
 *  +-----------------      .      -----------------+
 *  |                       .                       |
 *  +-----------------------------------------------+
 *
 *    : offset θ DNS message Ƭ
 */

typedef struct dns_rr {
					/* NAME			*/
	uint16_t	type;		/* TYPE			*/
	uint16_t	class;		/* CLASS		*/
	uint32_t	ttl;		/* TTL			*/
	uint16_t	rdlength;	/* RDATA length		*/
					/* RDATA		*/
	} T_DNS_RR;

#define T_DNS_RR_SIZE	(sizeof(uint16_t)*3+sizeof(uint32_t))

/*
 *  Question section 
 *
 *    0                    7  8                   15
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                       .                       |
 *  +-----------------      .      -----------------+
 *  |                     QNAME                     |
 *  +-----------------      .      -----------------+
 *  |                       .                       |
 *  +-----------------------------------------------+
 *  |                     QTYPE                     |
 *  +-----------------------------------------------+
 *  |                     QCLASS                    |
 *  +-----------------------------------------------+
 */

typedef struct dns_qd {
					/* QNAME		*/
	uint16_t	type;		/* TYPE			*/
	uint16_t	class;		/* CLASS		*/
	} T_DNS_QD;

#define T_DNS_QD_SIZE	(sizeof(uint16_t)*2)

/*
 *  SOA RDATA 
 *
 *    0                    7  8                   15
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                       .                       |
 *  +-----------------      .      -----------------+
 *  |                     MNAME                     |
 *  +-----------------      .      -----------------+
 *  |                       .                       |
 *  +-----------------------------------------------+
 *  |                       .                       |
 *  +-----------------      .      -----------------+
 *  |                     RNAME                     |
 *  +-----------------      .      -----------------+
 *  |                       .                       |
 *  +-----------------------------------------------+
 *  |                    SERIAL                     |
 *  |                                               |
 *  +-----------------------------------------------+
 *  |                    REFRESH                    |
 *  |                                               |
 *  +-----------------------------------------------+
 *  |                     RETRY                     |
 *  |                                               |
 *  +-----------------------------------------------+
 *  |                    EXPIRE                     |
 *  |                                               |
 *  +-----------------------------------------------+
 *  |                    MINIMUM                    |
 *  |                                               |
 *  +-----------------------------------------------+
 *
 *    : offset θ DNS message Ƭ
 */

typedef struct dns_soa {
					/* MNAME		*/
					/* RNAME		*/
	uint32_t	serial;		/* SERIAL		*/
	uint32_t	refresh;	/* REFRESH		*/
	uint32_t	retry;		/* RETRY		*/
	uint32_t	expire;		/* EXPIRE		*/
	uint32_t	minimum;	/* MINIMUM		*/
	} T_DNS_SOA;

#define T_DNS_SOA_SIZE	(sizeof(uint32_t)*5)

/*
 *  DNS 
 */

#define DNS_LABEL_LENGTH	63
#define DNS_NAME_LENGTH		255
#define DNS_UDP_MSG_LENGTH	512

#define DNS_MSG_COMP_MARK	0xc0

#endif	/* of #ifndef DNS_H */
