/*
 *  TINET (TCP/IP Protocol Stack)
 * 
 *  Copyright (C) 2001-2017 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: resolver.c 1.7 2017/6/1 8:49:57 abe $
 */

/*
 *  ۥ̾IP ɥ쥹Ѵ
 */

#include <string.h>

#ifdef TARGET_KERNEL_ASP

#include <kernel.h>
#include <t_syslog.h>
#include <syssvc/logtask.h>
#include "kernel_cfg.h"
#include "tinet_cfg.h"

#endif	/* of #ifdef TARGET_KERNEL_ASP */

#ifdef TARGET_KERNEL_JSP

#include <s_services.h>
#include <t_services.h>
#include "kernel_id.h"
#include "tinet_id.h"

#endif	/* of #ifdef TARGET_KERNEL_JSP */

#include <tinet_config.h>

#include <netinet/in.h>
#include <netinet/in_itron.h>

#include <netapp/netapp.h>
#include <netapp/netapp_var.h>
#include <netapp/resolver.h>

#if defined(USE_RESOLVER)

/*
 *  ɽѿ
 */

static uint8_t	dns_domain_name[DNS_NAME_LENGTH + 1] = RSLV_CFG_DNS_DOMAIN_NAME_STR;

#if defined(SUPPORT_INET6)
static T_IN6_ADDR in6_addr_dns      = IPV6_ADDR_DNS_INIT;
static T_IN6_ADDR in6_addr_dns_init = IPV6_ADDR_DNS_INIT;
#endif

#if defined(SUPPORT_INET4)
static T_IN4_ADDR in4_addr_dns = IPV4_ADDR_DNS;
#endif

static uint16_t	dns_msg_id = 0;

/*
 *  get_dns_header -- DNS å DNS إåФ
 */

static ER_UINT
get_dns_header (uint8_t *msg, uint_t length, T_RSLV_DNS_MSG *rslv)
{
	T_DNS_HDR	*dns_hdr;

	/* DNS إåФ*/
	if (length < sizeof(rslv->dns_hdr))
		return EV_RSLV_ESIZE;
	dns_hdr = (T_DNS_HDR*)msg;
	rslv->dns_hdr.id	= MSG2HOST16(dns_hdr->id);
	rslv->dns_hdr.code	= MSG2HOST16(dns_hdr->code);
	rslv->dns_hdr.qdcount	= MSG2HOST16(dns_hdr->qdcount);
	rslv->dns_hdr.ancount	= MSG2HOST16(dns_hdr->ancount);
	rslv->dns_hdr.nscount	= MSG2HOST16(dns_hdr->nscount);
	rslv->dns_hdr.arcount	= MSG2HOST16(dns_hdr->arcount);
	return sizeof(rslv->dns_hdr);
	}

/*
 *  dns_strtype -- DNS  type ʸ֤
 */

static const char *
dns_flag_qtype (uint16_t flags)
{
	if      (flags & DNS_LUP_FLAGS_QTYPE_AAAA)
		return "AAAA";
	else if (flags & DNS_LUP_FLAGS_QTYPE_A)
		return "   A";
	else if (flags & DNS_LUP_FLAGS_QTYPE_PTR)
		return " PTR";
	else
		return "unknown type";
	}

#if defined(SUPPORT_INET6)

/*
 *  udp6_dns -- UDP/IPv6 ˤ DNS Фå롣
 */

static ER
udp6_dns (uint16_t flags, T_IN6_ADDR *addr, uint8_t *msg, uint_t msg_size, uint_t length)
{
	T_IPV6EP	ep_dns;
	ER_UINT		len = E_OK;
	int		retry;

	/* DNS Ф IPv6 ɥ쥹̤λϥѥ᡼顼 */
	if (IN6_IS_ADDR_UNSPECIFIED(addr) || (in6_is_addr_ipv4mapped(addr) && addr->s6_addr32[3] == IPV4_ADDRANY)) {
		syslog(LOG_NOTICE, "[RSLV] undefined IPv6 DNS server address.");
		return E_PAR;
		}

	memcpy(&ep_dns.ipaddr, addr, sizeof(*addr));
	ep_dns.portno = RSLV_CFG_DNS_PORTNO;
	for (retry = NUM_DNS_UDP_RETRY + 1; retry -- > 0; ) {
		if ((len = udp6_snd_dat(UDP6_RESOLVER_CEPID, &ep_dns, msg, length, TMO_DNS_UDP_SND)) >= 0) {
			if ((len = udp6_rcv_dat(UDP6_RESOLVER_CEPID, &ep_dns, msg, msg_size, TMO_DNS_UDP_RCV)) >= 0)
				return len;
			}
		if (flags & DNS_LUP_FLAGS_MSG) {
			if (len == E_TMOUT)
				syslog(LOG_NOTICE, "[RSLV] TMOUT: %d, type: %s, server: %s.%d.",
				       NUM_DNS_UDP_RETRY + 1 - retry,
				       dns_flag_qtype(flags),
				       ipv62str(NULL, &ep_dns.ipaddr), ep_dns.portno);
			else {
				syslog(LOG_NOTICE, "[RSLV] error:  %s, type: %s,\n       server: %s.%d.",
				       itron_strerror(len),
				       dns_flag_qtype(flags),
				       ipv62str(NULL, &ep_dns.ipaddr), ep_dns.portno);
				return len;
				}
			}
		}
	return E_TMOUT;
	}

#endif	/* of #if defined(SUPPORT_INET6) */

#if defined(SUPPORT_INET4)

/*
 *  udp4_dns -- UDP/IPv4 ˤ DNS Фå롣
 */

static ER
udp4_dns (uint16_t flags, T_IN4_ADDR *addr, uint8_t *msg, uint_t msg_size, uint_t length)
{
	T_IPV4EP	ep_dns;
	ER_UINT		len = E_OK;
	int		retry;

	/* DNS Ф IPv4 ɥ쥹̤λϥѥ᡼顼 */
	if (*addr == IPV4_ADDRANY) {
		syslog(LOG_NOTICE, "[RSLV] undefined IPv4 DNS server address.");
		return E_PAR;
		}

	memcpy(&ep_dns.ipaddr, addr, sizeof(*addr));
	ep_dns.portno = RSLV_CFG_DNS_PORTNO;
	for (retry = NUM_DNS_UDP_RETRY + 1; retry -- > 0; ) {
		if ((len = udp_snd_dat(UDP4_RESOLVER_CEPID, &ep_dns, msg, length, TMO_DNS_UDP_SND)) >= 0) {
			if ((len = udp_rcv_dat(UDP4_RESOLVER_CEPID, &ep_dns, msg, msg_size, TMO_DNS_UDP_RCV)) >= 0)
				return len;
			}
		if (flags & DNS_LUP_FLAGS_MSG) {
			if (len == E_TMOUT)
				syslog(LOG_NOTICE, "[RSLV] TMOUT: %d, type: %s, server: %s.%d.",
				       NUM_DNS_UDP_RETRY + 1 - retry,
				       dns_flag_qtype(flags),
				       ip2str(NULL, &ep_dns.ipaddr), ep_dns.portno);
			else {
				syslog(LOG_NOTICE, "[RSLV] error: %s, type: %s, server: %s.%d.",
				       itron_strerror(len),
				       dns_flag_qtype(flags),
				       ip2str(NULL, &ep_dns.ipaddr), ep_dns.portno);
				return len;
				}
			}
		}
	return E_TMOUT;
	}

#endif	/* of #if defined(SUPPORT_INET4) */

/*
 *  skip_label -- ٥ɤФ
 */

static char *
skip_label (char *name)
{
	while (!(*name == '\0' || *name == '.'))
		name ++;
	return name;
	}

/*
 *  add_name -- å̾ɲä롣
 */

static ER_UINT
add_name (uint8_t *msg, uint_t size, uint_t offset, char *name)
{
	uint_t	label_len;
	char	*tail;

	while (*name != '\0') {

		/* åãƤ饨顼 */
		if (offset + 1 > size)
			return E_BOVR;

		/* ٥ν򸫤Ĥ롣*/
		tail = skip_label(name);

		/* ٥Ĺꤹ롣*/
		label_len = tail - name;
		if (label_len > DNS_LABEL_LENGTH) {	/* 63 ƥåȤĶȥ顼 */
			syslog(LOG_NOTICE, "[RSLV] label length(%d) too long > %d.",
			                   label_len, DNS_LABEL_LENGTH);
			return E_PAR;
			}

		*(msg + offset) = (uint8_t)label_len;
		offset ++;

		/* ٥̾ꤹ롣*/
		if (offset + label_len > size) {	/* åĶȥ顼 */
			syslog(LOG_NOTICE, "[RSLV] message length(%d) too long > %d.",
			                   offset + label_len, size);
			return E_BOVR;
			}

		memcpy((void*)(msg + offset), (void*)name, label_len);
		offset += label_len;

		/* Υ٥˰ư롣*/
		if (*tail == '\0')
			name = tail;
		else
			name = tail + 1;	/* '.' 򥹥åפ롣*/
		}

	return offset;
	}

/*
 *  add_ipv6addr -- å IPv6 ɥ쥹ɲä롣
 */

static ER_UINT
add_ipv6addr (uint8_t *msg, uint_t size, uint_t offset, char *name)
{
	T_IN6_ADDR	addr;
	ER_UINT		error;
	uint_t		len;
	int			col;
	char		digit[2];

	len = strlen(name);
	if (len == 0)
		return E_PAR;

	digit[1] = '\0';
	get_ipv6addr(&addr, name);
	for (col = sizeof(T_IN6_ADDR) * 2; -- col >= 0; ) {

		/* 16ʿ 1ͤ̾Ȥɲä롣*/
		digit[0] = (addr.s6_addr8[col / 2] >> ((1 - (col % 2)) << 2)) & 0x0f;
		if (digit[0] >= 10)
			digit[0] += 'a' - 10;
		else
			digit[0] += '0';
		if ((error = add_name(msg, size, offset, digit)) < 0)
			return error;
		offset = error;
		}

	/* "ip6" ɲä롣*/
	if ((error = add_name(msg, size, offset, "ip6")) < 0)
		return error;

	/* "arpa" ɲä롣*/
	return add_name(msg, size, error, "arpa");
	}

/*
 *  add_ipv4addr -- å IPv4 ɥ쥹ɲä롣
 */

static ER_UINT
add_ipv4addr (uint8_t *msg, uint_t size, uint_t offset, char *name)
{
	T_IN4_ADDR	addr;
	ER_UINT		error;
	uint_t		len, octet;
	int			col;
	char		*op, ostr[4];

	len = strlen(name);
	if (len == 0)
		return E_PAR;

	ostr[3] = '\0';
	get_ipv4addr(&addr, name);
	for (col = 0; col < sizeof(T_IN4_ADDR); col ++) {

		/* 1ƥåȤͤͤƥѴ롣*/
		octet = (addr >> (col << 3)) & 0xff;
		op = &ostr[3];
		while (octet > 0) {
			*(-- op) = octet % 10 + '0';
			octet   /= 10;
			}

		/* 1ƥåȤͤ̾Ȥɲä롣*/
		if ((error = add_name(msg, size, offset, op)) < 0)
			return error;
		offset = error;
		}

	/* "in-addr" ɲä롣*/
	if ((error = add_name(msg, size, offset, "in-addr")) < 0)
		return error;

	/* "arpa" ɲä롣*/
	return add_name(msg, size, error, "arpa");
	}

/*
 *  setup_dns_msg -- DNS å롣
 */

static ER
setup_dns_msg (uint16_t flags, char *name, uint8_t *msg, uint_t msg_size)
{
	T_DNS_HDR	dns_hdr;
	ER_UINT		offset;

	memset((void*)&dns_hdr, 0, sizeof(dns_hdr));

	dns_msg_id++;
	HOST2MSG16(dns_hdr.id,      dns_msg_id);
	HOST2MSG16(dns_hdr.code,    (flags & DNS_OPCODE_MASK) | DNS_RECURSION_DESIRED);
	HOST2MSG16(dns_hdr.qdcount, 1);

	memcpy((void*)msg, (void*)&dns_hdr, sizeof(dns_hdr));
	offset = sizeof(dns_hdr);

	switch (flags & DNS_LUP_FLAGS_NAME_MASK) {

	case DNS_LUP_FLAGS_NAME_HOST:
	case DNS_LUP_FLAGS_NAME_FQDN:

		/* ۥ̾ޤ FQDN λν*/
		/* ۥ̾ɲä롣*/
		if ((offset = add_name(msg, msg_size, offset, name)) < 0)
			return offset;

		if (offset - sizeof(dns_hdr) > DNS_NAME_LENGTH) {		/* ̾ 255 ƥåȤĶȥ顼 */
			syslog(LOG_NOTICE, "[RSLV] name length(%d) too long > %d.",
			                   offset - sizeof(dns_hdr), DNS_NAME_LENGTH);
			return E_PAR;
			}

		/* ۥ̾λϡɥᥤ̾ɲä롣*/
		if ((flags & DNS_LUP_FLAGS_NAME_MASK) == DNS_LUP_FLAGS_NAME_HOST) {

			/* ɥᥤ̾ɲä롣*/
			if ((offset = add_name(msg, msg_size, offset, (char *)dns_domain_name)) < 0)
				return offset;

			if (offset - sizeof(dns_hdr) > DNS_NAME_LENGTH) {		/* ̾ 255 ƥåȤĶȥ顼 */
				syslog(LOG_NOTICE, "[RSLV] name length(%d) too long > %d.",
				                   offset - sizeof(dns_hdr), DNS_NAME_LENGTH);
				return E_PAR;
				}
			}
		break;

	case DNS_LUP_FLAGS_NAME_IPV4:
		if ((offset = add_ipv4addr(msg, msg_size, offset, name)) < 0)
			return offset;
		break;

	case DNS_LUP_FLAGS_NAME_IPV6:
		if ((offset = add_ipv6addr(msg, msg_size, offset, name)) < 0)
			return offset;
		break;

	default:
		syslog(LOG_NOTICE, "[RSLV] unknown name type: %04x.", flags & DNS_LUP_FLAGS_NAME_MASK);
		return E_PAR;
		break;

		}

	/* ̾νλɡTYPECLASS ɲäƤåĹĶʤȤǧ롣*/
	if (offset + sizeof(uint16_t) * 2 + 1 > msg_size) {
		syslog(LOG_NOTICE, "[RSLV] message length(%d) too long > %d.",
		                   offset + sizeof(uint16_t) * 2 + 1, msg_size);
		return E_BOVR;
		}

	/* ̾νλɤꤹ롣*/
	*(msg + offset) = '\0';
	offset ++;

	/* DNS  TYPE ꤹ롣*/
	if (flags & DNS_LUP_FLAGS_QTYPE_PTR)
		host2msg16(msg + offset, DNS_TYPE_PTR);
	else if (flags & DNS_LUP_FLAGS_QTYPE_AAAA)
		host2msg16(msg + offset, DNS_TYPE_AAAA);
	else if (flags & DNS_LUP_FLAGS_QTYPE_A)
		host2msg16(msg + offset, DNS_TYPE_A);
	else {
		syslog(LOG_NOTICE, "[RSLV] unknown query type: %04x.", flags & DNS_LUP_FLAGS_QTYPE_MASK);
		return E_PAR;
		}
	offset += sizeof(uint16_t);

	/* DNS  CLASS ꤹ롣*/
	host2msg16(msg + offset, DNS_CLASS_IN);

	return offset + sizeof(uint16_t);
	}

/*
 *  skip_dns_name -- DNS å̾ɤФ
 *
 *    : ̷θ롣 2ӥåȤ 11ʤ鰵̷
 */

static ER_UINT
skip_dns_name (uint8_t *name, uint_t offset, uint_t length)
{
	while (*(name + offset)) {
		if ((*(name + offset) & 0xc0) == 0xc0) {
			if (offset + 2 > length)
				return EV_RSLV_ESIZE;
			else
				return offset + 2;
			}
		else if (offset + *name + 1 > length)
			return EV_RSLV_ESIZE;
		else
			offset += *name + 1;
		}
	return offset + 1;
	}

/*
 *  skip_dns_qd -- DNS Question section ɤФ
 */

static ER_UINT
skip_dns_qd (uint8_t *msg, uint_t offset, uint_t length, uint_t count)
{
	ER_UINT	next;

	while (count != 0) {
		/* name ɤФ*/
		if ((next = skip_dns_name(msg, offset, length)) < 0)
			return next;
		offset = next;

		/* typeclass ɤФ*/
		if ((offset + sizeof(uint16_t) + sizeof(uint16_t)) > length)
			return EV_RSLV_ESIZE;
		offset += sizeof(uint16_t) + sizeof(uint16_t);

		count --;
		}

	return offset;
	}

/*
 *  skip_dns_rr -- DNS RR ɤФ
 */

static ER_UINT
skip_dns_rr (uint8_t *msg, uint_t offset, uint_t length, uint_t count)
{
	ER_UINT		next;
	uint16_t	len = 0;

	while (count != 0) {
		/* name ɤФ*/
		if ((next = skip_dns_name(msg, offset, length)) < 0)
			return next;
		offset = next;

		/* typeclassTTL ɤФ*/
		if (offset + sizeof(uint16_t) + sizeof(uint16_t) + sizeof(uint32_t) > length)
			return EV_RSLV_ESIZE;
		offset += sizeof(uint16_t) + sizeof(uint16_t) + sizeof(uint32_t);

		/* RD length  RDATA ɤФ*/
		if (offset + sizeof(uint16_t) > length)
			return EV_RSLV_ESIZE;
		len = msg2host16(msg + offset);

		if (offset + sizeof(uint16_t) + len > length)
			return EV_RSLV_ESIZE;
		offset += sizeof(uint16_t) + len;

		count --;
		}

	return offset;
	}

/*
 *  dns_analyze_qd -- DNS å question section Ϥ롣
 */

ER_UINT
dns_analyze_qd (T_RSLV_QD *qd, uint_t offset, uint8_t *msg, uint_t length)
{
	ER_UINT		error;
	T_DNS_QD	*dns_qd;

	if ((error = skip_dns_name(msg, offset, length)) < 0)
		return error;
	offset = error;

	/* DNS åηΥǥå롣*/
	if (offset + T_DNS_QD_SIZE > length)
		return EV_RSLV_ESIZE;

	/*
	 *  DNS å SOA RDATA ХȥĴ
	 *  ¤Τ˥ԡ롣
	 */
	dns_qd = (T_DNS_QD*)(msg + offset);
	qd->type  = MSG2HOST16(dns_qd->type);
	qd->class = MSG2HOST16(dns_qd->class);

	return offset + T_DNS_QD_SIZE;
	}

/*
 *  dns_analyze_soa -- DNS å SOA RDATA Ϥ롣
 */

ER_UINT
dns_analyze_soa (T_RSLV_SOA *soa, uint_t offset, uint8_t *msg, uint_t length)
{
	ER_UINT		error;
	T_DNS_SOA	*dns_soa;

	/* MNAME 򥹥åפ롣*/
	if ((error = skip_dns_name(msg, offset, length)) < 0)
		return error;
	offset = error;

	/* RNAME 򥹥åפ롣*/
	if ((error = skip_dns_name(msg, offset, length)) < 0)
		return error;
	offset = error;

	/* DNS åηΥǥå롣*/
	if (offset + T_DNS_SOA_SIZE > length)
		return EV_RSLV_ESIZE;

	/*
	 *  DNS å SOA RDATA ХȥĴ
	 *  ¤Τ˥ԡ롣
	 */
	dns_soa = (T_DNS_SOA*)(msg + offset);
	soa->serial  = MSG2HOST32(dns_soa->serial);
	soa->refresh = MSG2HOST32(dns_soa->refresh);
	soa->retry   = MSG2HOST32(dns_soa->retry);
	soa->expire  = MSG2HOST32(dns_soa->expire);
	soa->minimum = MSG2HOST32(dns_soa->minimum);

	return offset + T_DNS_SOA_SIZE;
	}

/*
 *  dns_analyze_rr -- DNS å AN/NS/AR section Ϥ롣
 */

ER_UINT
dns_analyze_rr (T_RSLV_RR *rr, uint_t offset, uint8_t *msg, uint_t length)
{
	ER_UINT		error;
	T_DNS_RR	*dns_rr;

	if ((error = skip_dns_name(msg, offset, length)) < 0)
		return error;
	offset = error;

	/* DNS åηΥǥå롣*/
	if (offset + T_DNS_RR_SIZE > length)
		return EV_RSLV_ESIZE;

	/*
	 *  DNS å SOA RDATA ХȥĴ
	 *  ¤Τ˥ԡ롣
	 */
	dns_rr = (T_DNS_RR*)(msg + offset);
	rr->type	= MSG2HOST16(dns_rr->type);
	rr->class	= MSG2HOST16(dns_rr->class);
	rr->ttl		= MSG2HOST32(dns_rr->ttl);
	rr->rdlength	= MSG2HOST16(dns_rr->rdlength);
	rr->rdata_offset= offset + sizeof(rr->type) + sizeof(rr->class)
	                         + sizeof(rr->ttl)  + sizeof(rr->rdlength);

	return offset + T_DNS_RR_SIZE + rr->rdlength;
	}

/*
 *  analyze_dns_msg -- DNS åϤ롣
 */

static ER_UINT
analyze_dns_msg (uint8_t *msg, uint_t length, T_RSLV_DNS_MSG *rslv)
{
	ER_UINT	offset;

	/* DNS إåФ*/
	if ((offset = get_dns_header(msg, length, rslv)) < 0)
		return offset;

	/* question section ɤФ*/
	rslv->qd_offset = offset;
	if ((offset = skip_dns_qd(msg, offset, length, rslv->dns_hdr.qdcount)) < 0)
		return offset;

	/* answer section ɤФ*/
	rslv->an_offset = offset;
	if ((offset = skip_dns_rr(msg, offset, length, rslv->dns_hdr.ancount)) < 0)
		return offset;

	/* authority records section ɤФ*/
	rslv->ns_offset = offset;
	if ((offset = skip_dns_rr(msg, offset, length, rslv->dns_hdr.nscount)) < 0)
		return offset;

	/* additional records  section ɤФ*/
	rslv->ar_offset = offset;
	if ((offset = skip_dns_rr(msg, offset, length, rslv->dns_hdr.arcount)) < 0)
		return offset;

	return E_OK;
	}

/*
 *  dns_lookup_host_sub -- DNS ˤۥȤξؿˡ
 */

static ER_UINT
dns_lookup_host_sub (uint16_t flags, char *hostname, uint8_t *msg, 
                     uint_t msg_size, T_RSLV_DNS_MSG *rslv)
{
#if defined(SUPPORT_INET6) && defined(SUPPORT_INET4)
	uint16_t	proto;
#endif
	ER_UINT		length;
	ER		error;

	if ((length = setup_dns_msg(flags, hostname, msg, msg_size)) < 0)
		return length;

#if defined(SUPPORT_INET6)

#if defined(SUPPORT_INET4)

	if (flags & DNS_LUP_FLAGS_PROTO_IPV6) {
		proto = DNS_LUP_FLAGS_PROTO_IPV6;
		if ((length = udp6_dns(flags, &in6_addr_dns, msg, msg_size, length)) < 0)
			return length;
		}
	else if (flags & DNS_LUP_FLAGS_PROTO_IPV4) {
		proto = DNS_LUP_FLAGS_PROTO_IPV4;
		if ((length = udp4_dns(flags, &in4_addr_dns, msg, msg_size, length)) < 0)
			return length;
		}
	else {
		syslog(LOG_NOTICE, "[RSLV] unknown lookup protocol: %04x.", flags & DNS_LUP_FLAGS_PROTO_MASK);
		return E_PAR;
		}

#else	/* of #if defined(SUPPORT_INET4) */

	if (flags & DNS_LUP_FLAGS_PROTO_IPV6) {
		if ((length = udp6_dns(flags, &in6_addr_dns, msg, msg_size, length)) < 0)
			return length;
		}
	else {
		syslog(LOG_NOTICE, "[RSLV] unknown lookup protocol: %04x.", flags & DNS_LUP_FLAGS_PROTO_MASK);
		return E_PAR;
		}

#endif	/* of #if defined(SUPPORT_INET4) */

#else	/* of #if defined(SUPPORT_INET6) */

#if defined(SUPPORT_INET4)

	if (flags & DNS_LUP_FLAGS_PROTO_IPV4) {
		if ((length = udp4_dns(flags, &in4_addr_dns, msg, msg_size, length)) < 0)
			return length;
		}
	else {
		syslog(LOG_NOTICE, "[RSLV] unknown lookup protocol: %04x.", flags & DNS_LUP_FLAGS_PROTO_MASK);
		return E_PAR;
		}

#endif	/* of #if defined(SUPPORT_INET4) */

#endif	/* of #if defined(SUPPORT_INET6) */

	if ((error = analyze_dns_msg(msg, length, rslv)) != E_OK)
		return error;
	else {
		if ((rslv->dns_hdr.ancount > 0) && (flags & DNS_LUP_FLAGS_MSG)) {

#if defined(SUPPORT_INET6)

#if defined(SUPPORT_INET4)

			if (proto & DNS_LUP_FLAGS_PROTO_IPV6) {
				syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
				       dns_flag_qtype(flags),
				       ipv62str(NULL, &in6_addr_dns),
				       RSLV_CFG_DNS_PORTNO);
				}
			else {
				syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
				       dns_flag_qtype(flags),
				       ip2str(NULL, &in4_addr_dns),
				       RSLV_CFG_DNS_PORTNO);
				}

#else	/* of #if defined(SUPPORT_INET4) */

			syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
			       dns_flag_qtype(flags),
			       ipv62str(NULL, &in6_addr_dns),
			       RSLV_CFG_DNS_PORTNO);

#endif	/* of #if defined(SUPPORT_INET4) */

#else	/* of #if defined(SUPPORT_INET6) */

#if defined(SUPPORT_INET4)

			syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
			       dns_flag_qtype(flags),
			       ip2str(NULL, &in4_addr_dns),
			       RSLV_CFG_DNS_PORTNO);

#endif	/* of #if defined(SUPPORT_INET4) */

#endif	/* of #if defined(SUPPORT_INET6) */

			}
		return length;
		}
	}

#if defined(SUPPORT_INET6) && defined(SUPPORT_INET4)

/*
 *  udp64_dns -- DNS/IPv6DNS/IPv4 ˤ DNS Фå롣
 */

static ER_UINT
udp64_dns (uint16_t flags, char *hostname, uint8_t *msg, 
           uint_t msg_size, T_RSLV_DNS_MSG *rslv)
{
	ER_UINT		length;
	ER		error;
	uint16_t	proto;

	/* DNS å롣*/
	if ((length = setup_dns_msg(flags, hostname, msg, msg_size)) < 0)
		return length;

	/* UDP/IPv6  DNS Ф˾Ȳ񤹤롣*/
	proto = DNS_LUP_FLAGS_PROTO_IPV6;
	if ((length = udp6_dns(flags, &in6_addr_dns, msg, msg_size, length)) < 0) {
		/*  UDP/IPv6  DNS ФؤξȲ񤬥顼ˤʤäϡUDP/IPv4  DNS Ф˾Ȳ񤹤롣*/
		if ((length = setup_dns_msg(flags, hostname, msg, msg_size)) < 0)
			return length;

		/* UDP/IPv4  DNS Ф˾Ȳ񤹤롣*/
		proto = DNS_LUP_FLAGS_PROTO_IPV4;
		if ((length = udp4_dns(flags, &in4_addr_dns, msg, msg_size, length)) < 0)
			return length;
		}

	/* DNS åϤ롣*/
	if ((error = analyze_dns_msg(msg, length, rslv)) != E_OK)
		return error;
	else {
		if ((rslv->dns_hdr.ancount > 0) && (flags & DNS_LUP_FLAGS_MSG)) {
			if (proto & DNS_LUP_FLAGS_PROTO_IPV6) {
				syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
				       dns_flag_qtype(flags),
				       ipv62str(NULL, &in6_addr_dns),
				       RSLV_CFG_DNS_PORTNO);
				}
			else {
				syslog(LOG_NOTICE, "[RSLV] reslv:    type: %s, server: %s.%d.",
				       dns_flag_qtype(flags),
				       ip2str(NULL, &in4_addr_dns),
				       RSLV_CFG_DNS_PORTNO);
				}
			}
		return length;
		}
	}

#endif	/* of #if defined(SUPPORT_INET6) && defined(SUPPORT_INET4) */

/*
 *  dns_lookup_host -- DNS ˤۥȤξ롣
 */

ER_UINT
dns_lookup_host (uint16_t flags, char *hostname, uint8_t *msg, 
              uint_t msg_size, T_RSLV_DNS_MSG *rslv)
{
	ER	error;

#if defined(SUPPORT_INET6) && defined(SUPPORT_INET4)

	if ((flags & DNS_LUP_FLAGS_PROTO_MASK) == (DNS_LUP_FLAGS_PROTO_IPV6 | DNS_LUP_FLAGS_PROTO_IPV4)) {

		/*
		 *  IPv6/IPv4 ξꤵ줿ν
		 *  ޤ type == AAAA  DNS Ф˾Ȳ񤹤롣
		 */
		if (flags & DNS_LUP_FLAGS_QTYPE_PTR) {
			if ((error = udp64_dns(flags, hostname, msg, msg_size, rslv)) < 0)
				return error;
			}
		else if (flags & DNS_LUP_FLAGS_QTYPE_AAAA) {
			if ((error = udp64_dns(flags & ~DNS_LUP_FLAGS_QTYPE_A, hostname, msg, msg_size, rslv)) < 0)
				return error;
			}
		else {
			HOST2MSG16(rslv->dns_hdr.ancount, 0);
			error = E_OK;
			}

		if (rslv->dns_hdr.ancount == 0) {

			/*
			 *  type == AAAA ǥɥ쥹褬Ǥʤäν
			 *  type == A  DNS Ф˾Ȳ񤹤롣
			 */
			if (flags & DNS_LUP_FLAGS_QTYPE_A)
				return udp64_dns(flags & ~DNS_LUP_FLAGS_QTYPE_AAAA, hostname, msg, msg_size, rslv);
			else {
				syslog(LOG_NOTICE, "[RSLV] unknown query type: %04x.", flags & DNS_LUP_FLAGS_QTYPE_MASK);
				return E_PAR;
				}
			}
		else
			return error;
		}

	if ((flags & DNS_LUP_FLAGS_QTYPE_MASK) == (DNS_LUP_FLAGS_QTYPE_AAAA | DNS_LUP_FLAGS_QTYPE_A)) {

		/*
		 *  type == AAAA  A ξꤵ줿ν
		 *  ޤ type == AAAA  DNS Ф˾Ȳ񤹤롣
		 */
		if ((error = dns_lookup_host_sub(flags & ~DNS_LUP_FLAGS_QTYPE_A, hostname, msg, msg_size, rslv)) < 0)
			return error;

		if (rslv->dns_hdr.ancount == 0) {

			/*
			 *  type == AAAA ǥɥ쥹褬Ǥʤäν
			 *  type == A  DNS Ф˾Ȳ񤹤롣
			 */
			return dns_lookup_host_sub(flags & ~DNS_LUP_FLAGS_QTYPE_AAAA, hostname, msg, msg_size, rslv);
			}
#if 0
		else {
			syslog(LOG_NOTICE, "[RSLV] unknown query type: %04x.", flags & DNS_LUP_FLAGS_QTYPE_MASK);
			return E_PAR;
			}
#endif
		}

	else {
		/* IPv6/IPv4 Τɤ餫ꤵ줿ν */
		error = dns_lookup_host_sub(flags, hostname, msg, msg_size, rslv);
		}

#else	/* of #if defined(SUPPORT_INET6) && defined(SUPPORT_INET4) */

	error = dns_lookup_host_sub(flags, hostname, msg, msg_size, rslv);

#endif	/* of #if defined(SUPPORT_INET6) && defined(SUPPORT_INET4) */

	return error;
	}

#if defined(SUPPORT_INET6)

/*
 *  dns_in6_set_dname -- DNS  ɥᥤ̾ꤹDHCPv6ѡˡ
 *
 *    : new  NULL ꤹȡ᤹ͤ
 */

const uint8_t *
dns_in6_set_dname (const uint8_t *new, uint_t len)
{
	uint8_t	*dst;
	int	label_len, name_len = DNS_NAME_LENGTH;

	if (new == NULL)
		strcpy((char *)dns_domain_name, RSLV_CFG_DNS_DOMAIN_NAME_STR);
	else {
		dst = dns_domain_name;
		while ((*new != 0) && (name_len > 0) && (len > 0)) {
			label_len = *new ++;
			len --;
			while ((name_len > 0) && (len > 0) && (label_len > 0)) {
				*dst ++ = *new ++;
				name_len  --;
				len       --;
				label_len --;
				}
			if ((*new != 0) && (name_len > 0)) {
				*dst ++ = '.';
				name_len --;
				}
			}
		*dst = '\0';
		}
	return dns_domain_name;
	}

/*
 *  dns_in6_get_dname -- DNS  ɥᥤ̾롣
 */

const uint8_t *
dns_in6_get_dname (void)
{
	return dns_domain_name;
	}

/*
 *  dns_in6_set_addr -- DNS Ф IPv6 ɥ쥹ꤹDHCPv6ѡˡ
 *
 *    : new  NULL ꤹȡ᤹ͤ
 */

T_IN6_ADDR *
dns_in6_set_addr (T_IN6_ADDR *new)
{
	if (new == NULL)
		in6_addr_dns = in6_addr_dns_init;
	else
		in6_addr_dns = *new;
	return &in6_addr_dns;
	}

/*
 *  dns_in6_get_addr -- DNS Ф IPv6 ɥ쥹롣
 */

T_IN6_ADDR *
dns_in6_get_addr (T_IN6_ADDR *dst)
{
	if (dst == NULL)
		return NULL;
	else {
		*dst = in6_addr_dns;
		return dst;
		}
	}

#endif	/* of #if defined(SUPPORT_INET6) */

#if defined(SUPPORT_INET4)

/*
 *  dns_in4_set_dname -- DNS  ɥᥤ̾ꤹDHCPv4ѡˡ
 *
 *    : new  NULL ꤹȡ᤹ͤ
 */

const uint8_t *
dns_in4_set_dname (const uint8_t *new, uint_t len)
{
	uint8_t	*dst;
	int	name_len = DNS_NAME_LENGTH;

	if (new == NULL)
		strcpy((char *)dns_domain_name, RSLV_CFG_DNS_DOMAIN_NAME_STR);
	else {
		dst = dns_domain_name;
		while ((name_len -- > 0) && (len -- > 0) && *new)
			*dst ++ = *new ++;
		*dst = '\0';
		}
	return dns_domain_name;
	}

/*
 *  dns_in4_get_dname -- DNS  ɥᥤ̾롣
 */

const uint8_t *
dns_in4_get_dname (void)
{
	return dns_domain_name;
	}

/*
 *  dns_in4_set_addr -- DNS Ф IPv4 ɥ쥹ꤹDHCPv4ѡˡ
 *
 *    : new  NULL ꤹȡ᤹ͤ
 */

T_IN4_ADDR *
dns_in4_set_addr (T_IN4_ADDR *new)
{
	if (new == NULL)
		in4_addr_dns = IPV4_ADDR_DNS;
	else
		in4_addr_dns = *new;
	return &in4_addr_dns;
	}

/*
 *  dns_in4_get_addr -- DNS Ф IPv4 ɥ쥹롣
 */

T_IN4_ADDR *
dns_in4_get_addr (T_IN4_ADDR *dst)
{
	if (dst == NULL)
		return NULL;
	else {
		*dst = in4_addr_dns;
		return dst;
		}
	}

#endif	/* of #if defined(SUPPORT_INET4) */

/*
 *  dns_host_addr -- ۥ̾ IP ɥ쥹Ѵ롣
 */

ER
dns_host_addr (uint16_t flags, char *hostname, T_IN_ADDR *addr)
{
	T_RSLV_DNS_MSG	rslv;
	T_RSLV_RR	rr;
	ER_UINT		length, offset;
	ER		error = E_OK;
	int		scount;
	uint8_t 	*msg;

#if defined(SUPPORT_INET6) && defined(SUPPORT_INET4)

	T_IN4_ADDR	addr4;

#endif	/* of #if defined(SUPPORT_INET6) && defined(SUPPORT_INET4) */

	if ((error = tget_mpf(MPF_RSLV_SRBUF, (void*)&msg, TMO_FEVR)) != E_OK)
		return error;

	if ((length = dns_lookup_host(flags, hostname, msg, DNS_UDP_MSG_LENGTH, &rslv)) < 0) {
		error = length;
		goto err_ret;
		}

	offset = rslv.an_offset;
	for (scount = 1; scount <= rslv.dns_hdr.ancount; scount ++) {
		if ((offset = dns_analyze_rr(&rr, offset, msg, length)) < 0) {
			error = offset;
			goto err_ret;
			}

#if defined(SUPPORT_INET6)

		if (rr.type == DNS_TYPE_AAAA && rr.class == DNS_CLASS_IN &&  rr.rdlength == sizeof(*addr)) {
			memcpy((void*)addr, (void*)(msg + rr.rdata_offset), sizeof(*addr));
			break;
			}

#if defined(SUPPORT_INET4)

		if (rr.type == DNS_TYPE_A && rr.class == DNS_CLASS_IN && rr.rdlength == sizeof(addr4)) {
			addr4 = msg2host32(msg + rr.rdata_offset);
			in6_make_ipv4mapped(addr, addr4);
			break;
			}

#endif	/* of #if defined(SUPPORT_INET4) */

#else	/* of #if defined(SUPPORT_INET6) */

#if defined(SUPPORT_INET4)

		if (rr.type == DNS_TYPE_A && rr.class == DNS_CLASS_IN && rr.rdlength == sizeof(*addr)) {
			*addr = msg2host32((msg + rr.rdata_offset));
			break;
			}

#endif	/* of #if defined(SUPPORT_INET4) */

#endif	/* of #if defined(SUPPORT_INET6) */

		}
	if (scount > rslv.dns_hdr.ancount)
		error = EV_RSLV_UNEXPCT;

err_ret:
	rel_mpf(MPF_RSLV_SRBUF, (void*)msg);
	return error;
	}

#endif	/* of #if defined(USE_RESOLVER) */
